import json
from typing import Any, Dict

import pytest
from _pytest.pathlib import Path


def pytest_addoption(parser):
    group = parser.getgroup("terminal reporting", "report-log plugin options")
    group.addoption(
        "--report-log",
        action="store",
        metavar="path",
        default=None,
        help="Path to line-based json objects of test session events.",
    )


def pytest_configure(config):
    report_log = config.option.report_log
    if report_log and not hasattr(config, "workerinput"):
        config._report_log_plugin = ReportLogPlugin(config, Path(report_log))
        config.pluginmanager.register(config._report_log_plugin)


def pytest_unconfigure(config):
    report_log_plugin = getattr(config, "_report_log_plugin", None)
    if report_log_plugin:
        report_log_plugin.close()
        del config._report_log_plugin


class ReportLogPlugin:
    def __init__(self, config, log_path: Path):
        self._config = config
        self._log_path = log_path

        log_path.parent.mkdir(parents=True, exist_ok=True)
        self._file = log_path.open("w", buffering=1, encoding="UTF-8")

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None

    def _write_json_data(self, data):
        try:
            json_data = json.dumps(data)
        except TypeError:
            data = cleanup_unserializable(data)
            json_data = json.dumps(data)
        self._file.write(json_data + "\n")
        self._file.flush()

    def pytest_sessionstart(self):
        data = {"pytest_version": pytest.__version__, "$report_type": "SessionStart"}
        self._write_json_data(data)

    def pytest_sessionfinish(self, exitstatus):
        data = {"exitstatus": exitstatus, "$report_type": "SessionFinish"}
        self._write_json_data(data)

    def pytest_runtest_logreport(self, report):
        data = self._config.hook.pytest_report_to_serializable(
            config=self._config, report=report
        )
        self._write_json_data(data)

    def pytest_collectreport(self, report):
        data = self._config.hook.pytest_report_to_serializable(
            config=self._config, report=report
        )
        self._write_json_data(data)

    def pytest_terminal_summary(self, terminalreporter):
        terminalreporter.write_sep(
            "-", "generated report log file: {}".format(self._log_path)
        )


def cleanup_unserializable(d: Dict[str, Any]) -> Dict[str, Any]:
    # Credit to https://github.com/pytest-dev/pytest-reportlog
    """Return new dict with entries that are not json serializable by their str()."""
    result = {}
    for k, v in d.items():
        try:
            json.dumps({k: v})
        except TypeError:
            v = str(v)
        result[k] = v
    return result



import logging
import os
import re
from pathlib import Path

import pytest
from _pytest.config import Config, ExitCode
from _pytest.main import Session
from _pytest.nodes import Item
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from _pytest.stash import Stash, StashKey
from _pytest.terminal import TerminalReporter
from strip_ansi import strip_ansi

from pytest_tally.classes import TallyReport, TallySession, TallyTest
from pytest_tally.utils import LocakbleJsonFileUtils

DEFAULT_FILE = Path(os.getcwd()) / "tally-data.json"
FLUSH_TIME = 0.05

pytest_tally_enabled = StashKey[bool]()
pytest_tally_json_file = StashKey[Path]()
pytest_tally_session = StashKey[TallySession]()

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


def check_tally_enabled(config: Config) -> bool:
    stash: Stash = config.stash
    stash["pytest_tally_enabled"] = (
        bool(config.option.tally) if hasattr(config.option, "tally") else False
    )
    return stash["pytest_tally_enabled"]


def pytest_addoption(parser) -> None:
    group = parser.getgroup("tally")
    group.addoption(
        "--tally",
        action="store_true",
        help=(
            "Enable the pytest-tally plugin. Writes live summary results data to a JSON"
            " file for consumption by a dashboard client."
        ),
    ),
    group.addoption(
        "--tally-file",
        action="store",
        default=DEFAULT_FILE,
        help=(
            "Specify the file path to write the pytest-tally data to. Defaults to"
            " tally-data.json in the current working directory."
        ),
    )


def pytest_cmdline_main(config: Config) -> None:
    # Define stash values here since this is one of the first pytest hooks to run in a sssion
    stash: Stash = config.stash
    stash["pytest_tally_enabled"] = (
        bool(config.option.tally) if hasattr(config.option, "tally") else False
    )
    if not stash["pytest_tally_enabled"]:
        return
    stash["pytest_tally_json_file"] = (
        Path(config.option.tally_file)
        if hasattr(config.option, "tally_file")
        else DEFAULT_FILE
    )

    pytest_tally_session = stash.get("pytest_tally_session", None)
    if not pytest_tally_session:
        stash["pytest_tally_session"] = TallySession(config=config)


def write_json_to_file(config: Config) -> None:
    stash: Stash = config.stash
    file_path = stash.get("pytest_tally_json_file", DEFAULT_FILE)
    os.makedirs(file_path.parent, exist_ok=True)
    session_data = stash["pytest_tally_session"].to_json()
    lock_utils = LocakbleJsonFileUtils(file_path=file_path)
    lock_utils.overwrite_json(session_data)


def pytest_sessionstart(session: Session) -> None:
    if not check_tally_enabled(session.config):
        return

    pytest_tally_session = session.config.stash["pytest_tally_session"]
    pytest_tally_session.timer.start()
    pytest_tally_session.session_started = True
    pytest_tally_session.session_duration = pytest_tally_session.timer.elapsed
    write_json_to_file(session.config)


def pytest_collection_finish(session: Session) -> None:
    if not check_tally_enabled(session.config):
        return

    pytest_tally_session = session.config.stash["pytest_tally_session"]
    pytest_tally_session.num_tests_to_run = len(session.items)
    pytest_tally_session.session_duration = pytest_tally_session.timer.elapsed
    write_json_to_file(session.config)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_setup(item: Item):
    if not check_tally_enabled(item.session.config):
        yield
        return

    tally_test = TallyTest(node_id=item.nodeid)
    tally_test.timer.reset()
    tally_test.timer.start()
    pytest_tally_session = item.session.config.stash["pytest_tally_session"]
    pytest_tally_session.tally_tests[item.nodeid] = tally_test

    if pytest_tally_session.num_tests_have_run == 0:
        write_json_to_file(item.session.config)
    pytest_tally_session.num_tests_have_run += 1
    yield


@pytest.hookimpl(trylast=True)  # do not remove!
def pytest_configure(config: Config) -> None:
    if not check_tally_enabled(config):
        # yield
        return

    assert config.pluginmanager.hasplugin("terminal")

    stash = config.stash
    pytest_tally_session = stash.get("pytest_tally_session", None)
    if not pytest_tally_session:
        stash["pytest_tally_session"] = TallySession(config=config)

    # This code exists solely to extract the single 'lastline' of the session for
    # display in the dashboard. It is a hacky way to do it, but it works.
    tr = config.pluginmanager.getplugin("terminalreporter")
    if tr is not None:
        oldwrite = tr._tw.write

        def tee_write(s, **kwargs):
            lastline_matcher = re.compile(r"^==.*in\s\d+.\d+s.*=+")
            oldwrite(s, **kwargs)
            match = re.search(lastline_matcher, s)
            if match:
                pytest_tally_session.lastline_ansi = match.string.replace(
                    "=", ""
                ).strip()
                pytest_tally_session.lastline = (
                    strip_ansi(match.string).replace("=", "").strip()
                )
                write_json_to_file(config)

        tr._tw.write = tee_write


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item: Item, call: CallInfo) -> None:
    if not check_tally_enabled(item.session.config):
        yield
        return  # do we need both yield and return?

    else:
        session_config = item.session.config
        pytest_tally_session = session_config.stash["pytest_tally_session"]

        r = yield
        report = r.get_result()
        if report.when == "teardown":
            try:
                pytest_tally_session.tally_tests[item.nodeid].timer.pause()
            except KeyError:
                logger.warning(f"Could not find tally test for node ID {item.nodeid}")
                return
            pytest_tally_session.session_duration = pytest_tally_session.timer.elapsed
            write_json_to_file(item.session.config)

        if report.when in ("setup", "teardown") and report.outcome == "failed":
            outcome = "error"
        elif hasattr(report, "wasxfail"):
            if report.outcome in ("passed", "failed"):
                outcome = "xpassed"
            elif report.outcome == "skipped":
                outcome = "xfailed"
        else:
            outcome = report.outcome

        tally_report = TallyReport(
            node_id=report.nodeid,
            when=report.when,
            outcome=report.outcome,
        )

        try:
            tally_test = pytest_tally_session.tally_tests[tally_report.node_id]
            tally_test.reports[tally_report.when] = tally_report
        except KeyError:
            logger.warning(
                f"Could not find tally test for node ID {tally_report.node_id}"
            )
            return

        if tally_test.test_outcome:
            tally_test.timer.pause()
            tally_test.test_duration = tally_test.timer.elapsed
            return

        if tally_report.when == "setup" and outcome in ["error", "skipped"]:
            tally_test.timer.pause()
            tally_test.test_duration = tally_test.timer.elapsed
            tally_test.test_outcome = outcome.capitalize()
            return

        if tally_report.when == "call":
            tally_test.test_outcome = outcome.capitalize()
            return


def pytest_sessionfinish(session: Session, exitstatus: ExitCode) -> None:
    # This called after whole test run finished, right before returning the exit status to the system.
    if not check_tally_enabled(session.config):
        return

    session_config = session.config
    pytest_tally_session = session_config.stash["pytest_tally_session"]

    pytest_tally_session.timer.pause()
    pytest_tally_session.session_duration = pytest_tally_session.timer.elapsed
    pytest_tally_session.session_finished = True
    write_json_to_file(session.config)



import argparse
import json
import logging
import os

from flask import Flask, jsonify, render_template

app = Flask(__name__)

# Global variables
results = None
fetch_rate = 10  # Default fetch rate in seconds


def read_json_file(file_path):
    global results
    with open(file_path) as file:
        results = json.load(file)


@app.route("/")
def index():
    read_json_file(app.config["JSON_FILE_PATH"])
    return render_template("index.html", results=results, fetch_rate=fetch_rate)


@app.route("/results")
def get_results():
    read_json_file(app.config["JSON_FILE_PATH"])
    return jsonify(results)


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Flask app with customizable JSON file path and debug mode."
    )
    parser.add_argument(
        "json_file", metavar="JSON_FILE", type=str, help="path to the JSON file"
    )
    parser.add_argument(
        "--port",
        metavar="PORT",
        type=int,
        default=8080,
        help="listening port for the app",
    )
    parser.add_argument("--debug", action="store_true", help="enable debug mode")
    parser.add_argument(
        "--log-level",
        metavar="LOG_LEVEL",
        type=str,
        default="ERROR",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        help="log level for Werkzeug",
    )
    parser.add_argument(
        "--fetch-rate",
        metavar="FETCH_RATE",
        type=int,
        default=2000,
        help="fetch rate (in ms) - effectively the update rate of the web app",
    )
    return parser.parse_args()


def configure_logging(log_level):
    log = logging.getLogger("werkzeug")
    log.setLevel(getattr(logging, log_level))


def main():
    default_json_file = os.path.join(os.getcwd(), "tally-data.json")

    # Parse command-line arguments
    parser = argparse.ArgumentParser(
        description="Flask app with customizable JSON file path and debug mode."
    )
    parser.add_argument(
        "json_file",
        metavar="JSON_FILE",
        type=str,
        nargs="?",
        default=default_json_file,
        help="path to the JSON file (default: %(default)s)",
    )
    parser.add_argument(
        "--port",
        metavar="PORT",
        type=int,
        default=8080,
        help="listening port for the app",
    )
    parser.add_argument("--debug", action="store_true", help="enable debug mode")
    parser.add_argument(
        "--log-level",
        metavar="LOG_LEVEL",
        type=str,
        default="ERROR",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        help="log level for Werkzeug",
    )
    parser.add_argument(
        "--fetch-rate",
        metavar="FETCH_RATE",
        type=int,
        default=2000,
        help="fetch rate (in ms) - effectively the update rate of the web app",
    )
    args = parser.parse_args()

    # Configure and run the Flask app
    app.config["JSON_FILE_PATH"] = args.json_file
    fetch_rate = args.fetch_rate
    print(fetch_rate)  # Keeping Flake8 happy for now
    configure_logging(args.log_level)
    print(f"Starting flask app with parameters: {args}")
    app.run(debug=args.debug, port=args.port)


if __name__ == "__main__":
    main()



import fcntl
import json
import os
from pathlib import Path
from typing import Any, Dict


def clear_file(filename: Path) -> None:
    with open(filename, "w") as jfile:
        jfile.write("")


class LocakbleJsonFileUtils:
    """
    Class to handle locking and reading/writing to a json file

    __init__ Args:
        file_path (Path): Path to the json file

    Public Methods:
        read_json: Read the json file and return the data
        write_json: Write the json file with the data

    Example:
        >>> file_path = Path("test.json")
        >>> clear_file(file_path)
        >>> utils = LocakbleJsonFileUtils(file_path)
        >>> utils.write_json({"test": "test"})
        >>> utils.read_json()
        => {'test': 'test'}
    """

    def __init__(self, file_path: Path):
        assert isinstance(file_path, Path), f"File {file_path} must be a Path object"
        assert file_path.exists(), f"File {file_path} must exist"
        assert file_path.suffix == ".json", f"File {file_path} must be a json file"
        self.file_path: Path = file_path
        self.file: Any = None

    def _acquire_read_lock(self):
        os.makedirs(self.file_path.parent, exist_ok=True)
        self.file = open(self.file_path, "r")
        fcntl.flock(self.file, fcntl.LOCK_SH)

    def _acquire_overwrite_lock(self):
        os.makedirs(self.file_path.parent, exist_ok=True)
        self.file = open(self.file_path, "w")
        fcntl.flock(self.file, fcntl.LOCK_EX)

    def _acquire_append_lock(self):
        os.makedirs(self.file_path.parent, exist_ok=True)
        self.file = open(self.file_path, "a")
        fcntl.flock(self.file, fcntl.LOCK_EX)

    def _release_lock(self):
        fcntl.flock(self.file, fcntl.LOCK_UN)
        self.file.close()
        self.file = None

    def read_json(self):
        self._acquire_read_lock()
        try:
            data = json.load(self.file)
        except json.decoder.JSONDecodeError:
            data = {}
        except FileNotFoundError:
            os.makedirs(self.file_path.parent, exist_ok=True)
            data = {}
        finally:
            self._release_lock()
        return data

    def overwrite_json(self, data: Dict[str, Any]):
        self._acquire_overwrite_lock()
        try:
            self.file.seek(0)
            json.dump(data, self.file, indent=4)
            self.file.truncate()
        finally:
            self._release_lock()

    def append_json(self, data: Dict[str, Any]):
        self._acquire_append_lock()
        try:
            self.file.seek(0)
            json.dump(data, self.file, indent=4)
            self.file.truncate()
        finally:
            self._release_lock()



from _pytest.config import Config
from count_timer import CountTimer


class TallyCountTimer(CountTimer):
    """
    Overriding CountTimer to add a to_json method. For this module all
    we care about is duration, and if the counter is running or not.
    """

    def __init__(self, duration=0):
        super().__init__(duration)

    def to_json(self):
        return {
            "elapsed": self.elapsed,
            "running": self.running,
            "finished": self.elapsed > 0 and not self.running,
        }


class TallySession:
    """
    Class to hold pertinent info for the entire Pytest test session.
    A test session is made up of one or more tests.
    """

    def __init__(
        self,
        config: Config,
        session_started: bool = False,
        session_finished: bool = False,
        num_tests_to_run: int = 0,
        num_tests_have_run: int = 0,
        session_duration: float = 0.0,
        lastline: str = "",
        lastline_ansi: str = "",
        timer: TallyCountTimer = TallyCountTimer(),
        tally_tests: dict = {},
    ) -> None:
        self.session_started = session_started
        self.session_finished = session_finished
        self.session_duration = session_duration
        self.num_tests_to_run = num_tests_to_run
        self.num_tests_have_run = num_tests_have_run
        self.timer = timer
        self.lastline = lastline
        self.lastline_ansi = lastline_ansi
        self.tally_tests = tally_tests
        self.config = config

    def to_json(self):
        return {
            "session_started": self.session_started,
            "session_finished": self.session_finished,
            "session_duration": self.session_duration,
            "num_tests_to_run": self.num_tests_to_run,
            "num_tests_have_run": self.num_tests_have_run,
            "timer": self.timer.to_json(),
            "lastline": self.lastline,
            "lastline_ansi": self.lastline_ansi,
            "tally_tests": {k: v.to_json() for k, v in self.tally_tests.items()},
        }


class TallyTest:
    """
    Class to hold pertinent info for each Pytest test executed.
    A test is made up of one or more test reports, usually one each
    corresponding to setup, call and teardown phases of the test.
    """

    def __init__(
        self,
        node_id: str = None,
        test_duration: float = 0.0,
        timer: TallyCountTimer = TallyCountTimer(),
        test_outcome: str = None,
        reports: dict = {},
    ) -> None:
        self.node_id = node_id
        self.test_duration = test_duration
        self.timer = timer
        self.test_outcome = test_outcome
        self.reports = reports

    def to_json(self):
        return {
            "node_id": self.node_id,
            "test_duration": self.test_duration,
            "test_outcome": self.test_outcome,
            "timer": self.timer.to_json(),
            "reports": {k: v.to_json() for k, v in self.reports.items()},
        }


class TallyReport:
    """
    Class to hold pertinent info for individual Pytest TestReport items.
    """

    def __init__(
        self,
        node_id: str = None,
        when: str = None,
        outcome: str = None,
    ) -> None:
        self.node_id = node_id
        self.when = when
        self.outcome = outcome

    def to_json(self):
        return {
            "node_id": self.node_id,
            "when": self.when,
            "outcome": self.outcome,
        }



import os
import time
from argparse import ArgumentParser, Namespace
from pathlib import Path
from threading import Event, Thread

from blessed import Terminal
from quantiphy import Quantity, render
from rich.box import ROUNDED as rounded
from rich.console import Console, Group, group
from rich.live import Live
from rich.panel import Panel
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn
from rich.status import Status
from rich.table import Table
from rich.text import Text

from pytest_tally import __version__
from pytest_tally.plugin import DEFAULT_FILE, TallySession
from pytest_tally.utils import LocakbleJsonFileUtils, clear_file

OUTCOME_STYLES = {
    "passed": "green",
    "failed": "bold red",
    "skipped": "yellow",
    "error": "bold magenta",
    "xfail": "red",
    "xpass": "bold yellow",
    "rerun": "yellow",
}


class Duration(Quantity):
    units = "s"
    prec = 2


class CmdLineOptions:
    def __init__(self, args: Namespace) -> None:
        self.filename = Path(args.filename) if args.filename else Path(DEFAULT_FILE)
        assert (
            self.filename.suffix == ".json"
        ), "'filename', if specified, must end in '.json'"

        self.max_rows = args.max_rows if hasattr(args, "max_rows") else 0
        self.lines = args.lines
        self.persist = args.persist if hasattr(args, "persist") else False


class Stats:
    def __init__(self, options: CmdLineOptions) -> None:
        self.options: CmdLineOptions = options
        self.tot_num_to_run: int = 0
        self.num_running: int = 0
        self.num_finished: int = 0
        self.testing_started: bool = False
        self.testing_complete: bool = False

    def _get_test_session_data(self, init: bool = False) -> TallySession:
        lock_utils = LocakbleJsonFileUtils(file_path=self.options.filename)
        if init:
            return TallySession(
                session_started=False,
                session_finished=False,
                session_duration=0.0,
                num_tests_to_run=0,
                num_tests_have_run=0,
                timer=None,
                lastline="",
                lastline_ansi="",
                tally_tests={},
                config=None,
            )
        j = lock_utils.read_json()
        if j:
            return TallySession(**j, config=None)

    def update_stats(self, init: bool = False) -> None:
        """Retrieve latest info from json file"""
        self.test_session_data = self._get_test_session_data(init=init)
        if self.test_session_data:
            self.tot_num_to_run = self.test_session_data.num_tests_to_run
            self.num_running = len(
                [
                    test
                    for test in self.test_session_data.tally_tests.values()
                    if test["timer"]["running"]
                ]
            )
            self.num_finished = len(
                [
                    test
                    for test in self.test_session_data.tally_tests.values()
                    if test["timer"]["finished"]
                ]
            )
            # self.testing_started = self.num_running > 0
            self.testing_started = self.test_session_data.session_started
            self.testing_complete = self.test_session_data.session_finished


class TallyApp:
    def __init__(self, args: Namespace):
        self.options = CmdLineOptions(args)
        self.stats = Stats(self.options)

        self.console = Console()
        self.term = Terminal()
        self.event = Event()
        self.table = Table(
            highlight=True, expand=True, show_lines=args.lines, box=rounded
        )

        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            expand=True,
        )
        self.task_id = self.progress.add_task(
            "Waiting for tests to start...", start=False
        )
        self.panel_progress = Panel(self.progress)

    def kb_input(self):
        while True:
            with self.term.cbreak():
                if self.event.is_set() and not self.options.persist:
                    os._exit(0)
                key = self.term.inkey(timeout=1).lower()
                if key and key == "q":
                    os._exit(0)

    def main_panel_group(self, stylize_last_line: bool = True) -> Group:
        # Main table (no panel container; it stands alone, looks better that way);
        # it's ok to loop over a Rich Table because it just redraws each iteration;
        # in fact, the Table class does not even have an update() method.
        self.table = Table(
            highlight=True, expand=True, show_lines=self.options.lines, box=rounded
        )
        self.table.add_column("Test")
        self.table.add_column("Duration")
        self.table.add_column("Outcome")

        # Wait till plugin.py starts to populate results file
        if (
            not hasattr(self.stats.test_session_data, "tally_tests")
            or not self.stats.test_session_data.tally_tests
        ):
            return self.panel_progress

        # Accommodate optional execution flag for max rows to display ("-x"),
        num_table_rows = (
            len(self.stats.test_session_data.tally_tests)
            if self.options.max_rows == 0
            else self.options.max_rows
        )
        tally_tests = list(self.stats.test_session_data.tally_tests.values())[
            -num_table_rows:
        ]

        # For each test result, add a row to the table with test info
        for _, test in enumerate(tally_tests):
            name = Text(test["node_id"], style="bold blue")
            duration = (
                Duration(str(test["test_duration"])) if test["test_duration"] else 0.0
            )
            outcome = (
                Text(test["test_outcome"]) if test["test_outcome"] else Text("---")
            )
            for key, value in OUTCOME_STYLES.items():
                if key in outcome.plain.lower():
                    outcome.stylize(value)
                    name.stylize(value)
                    break
            else:
                outcome.stylize("bold blue")
                name.stylize("bold blue")

            # Show spinny progress icon for last line of table while session running
            if stylize_last_line:
                if _ == len(tally_tests) - 1:
                    name = Status(name)
                    self.table.add_row(name, Status(""), outcome)
                else:
                    self.table.add_row(name, render(duration, "s"), outcome)
            else:
                self.table.add_row(name, render(duration, "s"), outcome)

        self.stats.update_stats()

        if self.stats.testing_started:
            if not self.stats.testing_complete:
                # self.progress.tasks[
                #     self.task_id
                # ].description = "Testing In Progress..."
                if not self.progress.tasks[self.task_id].started:
                    self.progress.start_task(self.task_id)
                self.progress.update(
                    self.task_id,
                    total=self.stats.tot_num_to_run,
                    completed=self.stats.num_finished,
                    description="Testing In Progress...",
                    refresh=True,
                )
            else:  # testing complete
                self.progress.update(
                    self.task_id,
                    total=self.stats.tot_num_to_run,
                    completed=self.stats.num_finished,
                    description="Testing Complete",
                    refresh=True,
                )
                time.sleep(1)
                self.progress.stop_task(self.task_id)

        # Rederable group - members depend on what phase of test session we are in:
        # Before tests start, only show progress bar
        # After tests start, show progress bar and table
        # After tests finish, show progress bar, table, and last line of test output
        @group()
        def get_panels(finished: bool = False):
            if not self.stats.testing_started:
                yield self.panel_progress
            elif self.stats.testing_started and not self.stats.testing_complete:
                yield self.table
                yield self.panel_progress
            elif self.stats.testing_started and self.stats.testing_complete:
                self.stats.update_stats()
                if self.stats.test_session_data and hasattr(
                    self.stats.test_session_data, "lastline_ansi"
                ):
                    last_line_ansi = self.stats.test_session_data.lastline_ansi
                else:
                    last_line_ansi = ""
                last_line = Text.from_ansi(last_line_ansi)
                yield self.table
                yield self.panel_progress
                yield Panel(last_line)

        return get_panels()

    def rich_client(self) -> None:
        # Put thread into a loop that proceseses pytest results until test
        # session is complete
        self.stats.update_stats(init=True)

        with Live(
            self.main_panel_group(),
            vertical_overflow="visible",
        ) as live:
            while not self.stats.testing_complete:
                live.update(self.main_panel_group())
                self.stats.update_stats()

            # Don't show spinny progress icon since tests are now finished,
            # otherwise it will appear frozen in time
            live.update(self.main_panel_group(stylize_last_line=False))

        # Make a final query of the stats to ensure latest results
        # Set the event, to signal to the kb_input thread to exit
        self.stats.update_stats()
        self.event.set()


def main():
    # CLI arguments
    parser = ArgumentParser(prog="tally")
    parser.add_argument("filename", nargs="?", help="path to data file")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-l",
        "--lines",
        action="store_true",
        default=False,
        help="draw separation [l]ines in between each table row (default: False)",
    )
    parser.add_argument(
        "-x",
        "--max_rows",
        action="store",
        type=int,
        default=0,
        help="ma[x] number of rows to display (default: 0 [no limit])",
    )
    parser.add_argument(
        "-f",
        "--file-path",
        action="store",
        type=str,
        default=Path.cwd() / "tally-data.json",
        help="ma[x] number of rows to display (default: 0 [no limit])",
    )
    args = parser.parse_args()

    # Create main app
    tally_app = TallyApp(args)
    tally_app.console.clear()
    clear_file(tally_app.options.filename)

    # Start threads and signaling event
    tally_app.event = Event()
    t1 = Thread(target=tally_app.rich_client)
    t2 = Thread(target=tally_app.kb_input)
    t1.start()
    t2.start()
    t1.join()
    t2.join()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import find_packages, setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding="utf-8").read()


setup(
    name="pytest-tally",
    version="1.3.0",
    author="Jeff Wright",
    author_email="jeff.washcloth@gmail.com",
    license="MIT",
    url="https://github.com/jeffwright13/pytest-tally",
    description=(
        "A Pytest plugin to generate realtime summary stats, and display them"
        " in-console using a text-based dashboard."
    ),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    py_modules=["pytest_tally"],
    python_requires=">=3.8",
    install_requires=[
        "blessed==1.20.0",
        "count-timer==0.3.8",
        "flask==2.3.2",
        "pytest>=6.2.5",
        "quantiphy==2.19",
        "rich==13.3.2",
        "single-source==0.3.0",
        "strip-ansi==0.1.1",
    ],
    setup_requires=["setuptools_scm"],
    include_package_data=True,
    classifiers=[
        "Framework :: Pytest",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Testing",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="pytest pytest-plugin testing tui rich blessed",
    entry_points={
        "pytest11": ["pytest_tally = pytest_tally.plugin"],
        "console_scripts": [
            "tally-rich = pytest_tally.clients.rich_dashboard:main",
            "tally-flask = pytest_tally.clients.app:main",
            "tally-tk = pytest_tally.clients.tk_client:main",
        ],
    },
)



import json
import logging
import shutil
import sys
import tkinter as tk
import tkinter.font as tkfont
from dataclasses import dataclass
from pathlib import Path
from tkinter import filedialog
from tkinter.ttk import Notebook, Progressbar

from quantiphy import Quantity, render
from watchdog.events import FileSystemEventHandler, LoggingEventHandler
from watchdog.observers import Observer

from pytest_tally import __version__
from pytest_tally.plugin import DEFAULT_FILE, TallySession
from pytest_tally.utils import LocakbleJsonFileUtils, clear_file

TERM_SIZE = shutil.get_terminal_size()
APP_HEIGHT = 700
APP_WIDTH = 1020
APP_TITLE = f"Pytest Tally v{__version__}"


@dataclass
class TableColumn:
    name: str
    width: int


TABLE_COLUMNS = [
    TableColumn("  node_id", 70),
    TableColumn("  duration", 20),
    TableColumn("  outcome", 20),
]


class Duration(Quantity):
    units = "s"
    prec = 2


class Stats:
    def __init__(self) -> None:
        self.tot_num_to_run: int = 0
        self.num_running: int = 0
        self.num_finished: int = 0
        self.testing_started: bool = False
        self.testing_complete: bool = False

    def _get_test_session_data(
        self, file_path: Path, init: bool = False
    ) -> TallySession:
        lock_utils = LocakbleJsonFileUtils(file_path=file_path)
        if init:
            return TallySession(
                session_started=False,
                session_finished=False,
                session_duration=0.0,
                num_tests_to_run=0,
                num_tests_have_run=0,
                timer=None,
                lastline="",
                lastline_ansi="",
                tally_tests={},
                config=None,
            )
        j = lock_utils.read_json()
        if j:
            return TallySession(**j, config=None)

    def update_stats(self, file_path: Path, init: bool = False) -> None:
        """Retrieve latest info from json file"""
        self.test_session_data = self._get_test_session_data(
            file_path=file_path, init=init
        )
        if self.test_session_data:
            self.tot_num_to_run = self.test_session_data.num_tests_to_run
            self.num_running = len(
                [
                    test
                    for test in self.test_session_data.tally_tests.values()
                    if test["timer"]["running"]
                ]
            )
            self.num_finished = len(
                [
                    test
                    for test in self.test_session_data.tally_tests.values()
                    if test["timer"]["finished"]
                ]
            )
            self.testing_started = self.test_session_data.session_started
            self.testing_complete = self.test_session_data.session_finished


class FileChangeEventHandler(FileSystemEventHandler):
    def __init__(self, callback):
        self.callback = callback

    def on_modified(self, event):
        if not event.is_directory and Path(event.src_path) == self.file_path:
            self.callback()


class TestResultsGUI:
    def __init__(self, root):
        self.root = root
        self.root.title(APP_TITLE)
        self.stats = Stats()
        self.file_path = DEFAULT_FILE
        self.max_rows = None

        self.create_widgets()
        self.file_observer = None  # Initialize the file_observer attribute
        if self.file_path:
            self.start_file_monitoring()

    def create_widgets(self):
        self.title_label = tk.Label(
            self.root, text="Test Results", font=("Arial", 18, "bold")
        )
        self.title_label.grid(row=0, column=0, columnspan=3, pady=10)

        self.resize_button = tk.Button(
            self.root, text="Resize", command=self.resize_window
        )
        self.resize_button.grid(row=1, column=0, pady=10)

        self.notebook = Notebook(self.root)
        self.notebook.grid(row=1, column=0, columnspan=3, pady=10, sticky="nsew")

        self.table_tab = tk.Frame(self.notebook)
        self.notebook.add(self.table_tab, text="Results")

        self.config_tab = tk.Frame(self.notebook)
        self.notebook.add(self.config_tab, text="Configuration")

        self.create_config_widgets()
        self.create_table_widgets()

        # Create a label to display the lastline
        self.lastline_label = tk.Label(
            self.root,
            font=tkfont.Font(family="Arial", size=14, weight="bold"),
            justify=tk.CENTER,
            # anchor="nsew",
            wraplength=APP_WIDTH - 100,
        )
        self.lastline_label.grid(
            row=2, column=0, columnspan=3, padx=10, pady=5, sticky="nsew"
        )

    def create_config_widgets(self):
        self.config_frame = tk.Frame(self.config_tab)
        self.config_frame.pack(pady=10)

        self.file_label = tk.Label(self.config_frame, text="File Path:")
        self.file_label.grid(row=0, column=0, sticky="w")

        self.file_entry = tk.Entry(self.config_frame, width=20)
        self.file_entry.grid(row=0, column=1, padx=5, sticky="w")

        self.browse_button = tk.Button(
            self.config_frame, text="Browse", command=self.browse_file
        )
        self.browse_button.grid(row=0, column=2, padx=5, sticky="w")

        self.max_rows_label = tk.Label(self.config_frame, text="Max Rows:")
        self.max_rows_label.grid(row=0, column=3, padx=5, sticky="w")

        self.max_rows_entry = tk.Entry(self.config_frame, width=10)
        self.max_rows_entry.grid(row=0, column=4, padx=5, sticky="w")

        self.apply_button = tk.Button(
            self.config_frame, text="Apply", command=self.apply_config
        )
        self.apply_button.grid(row=0, column=5, padx=5, sticky="w")

    def create_table_widgets(self):
        self.table_frame = tk.Frame(self.table_tab)
        self.table_frame.pack(pady=10)

        self.table_header = tk.Frame(self.table_frame)
        self.table_header.pack()

        for i, header in enumerate(TABLE_COLUMNS):
            label = tk.Label(
                self.table_header,
                text=header.name,
                font=("Arial", 14, "bold"),
                width=header.width,
                anchor="w",
            )
            label.grid(row=0, column=i, padx=10, pady=5, sticky="w")

        self.table_canvas = tk.Canvas(self.table_frame, height=400)
        self.table_canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        # Create the table body and configure the canvas
        self.table_body = tk.Frame(self.table_canvas)
        self.table_canvas.create_window((0, 0), window=self.table_body, anchor="nw")

        # Create the scrollbar after creating the canvas
        self.table_scrollbar = tk.Scrollbar(
            self.table_frame, orient=tk.VERTICAL, command=self.table_canvas.yview
        )
        self.table_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        # Configure the canvas to use the scrollbar
        self.table_canvas.configure(yscrollcommand=self.table_scrollbar.set)

        # Bind the scrollbar and canvas together
        self.table_scrollbar.config(command=self.table_canvas.yview)
        self.table_canvas.bind_all(
            "<MouseWheel>",
            lambda event: self.table_canvas.yview_scroll(
                int(-1 * (event.delta / 120)), "units"
            ),
        )

        # Configure the table_body to expand with the window size
        self.table_body.bind(
            "<Configure>",
            lambda event: self.table_canvas.configure(
                scrollregion=self.table_canvas.bbox("all")
            ),
        )

        # Attach the table_body to the canvas
        self.table_canvas.create_window((0, 0), window=self.table_body, anchor="nw")

        # Create the footer frame
        self.table_footer = tk.Frame(self.table_frame)
        self.table_footer.pack(side=tk.BOTTOM, fill=tk.X, pady=10)

    def browse_file(self):
        file_path = filedialog.askopenfilename(filetypes=[("JSON files", "*.json")])
        self.file_entry.delete(0, tk.END)
        self.file_entry.insert(0, file_path)

    def apply_config(self):
        self.file_path = Path(self.file_entry.get())
        if self.file_path:
            self.start_file_monitoring()

        max_rows = self.max_rows_entry.get()

        if self.file_path and max_rows:
            self.max_rows = int(max_rows)

            self.fetch_results()

    def fetch_results(self):
        if self.file_path is not None and self.file_path.is_file():
            self.stats.update_stats(file_path=self.file_path)

            if self.stats.test_session_data:
                self.update_table(self.stats.test_session_data.tally_tests)
            else:
                print("Error loading test session data.")
        else:
            print("Invalid file path.")

    def update_table(self, results):
        # Clear the table body
        for widget in self.table_body.winfo_children():
            widget.destroy()

        # Sort the results by node_id
        sorted_results = sorted(results.values(), key=lambda x: x["node_id"])

        # Display the maximum number of rows specified
        for i, result in enumerate(sorted_results[: self.max_rows]):
            row_frame = tk.Frame(self.table_body)
            row_frame.grid(row=i, column=0, padx=5, pady=2, sticky="w")

            node_id_label = tk.Label(
                row_frame,
                text=result["node_id"],
                font=("Arial", 14),
                width=TABLE_COLUMNS[0].width,
                anchor="w",
            )
            node_id_label.grid(row=0, column=0, padx=10, pady=5, sticky="w")

            duration_label = tk.Label(
                row_frame,
                text=render(Duration(result["test_duration"]), "s"),
                font=("Arial", 14),
                width=TABLE_COLUMNS[1].width,
                anchor="w",
            )
            duration_label.grid(row=0, column=1, padx=10, pady=5, sticky="w")

            outcome_label = tk.Label(
                row_frame,
                text=result["test_outcome"],
                font=("Arial", 14),
                width=TABLE_COLUMNS[2].width,
                anchor="w",
            )
            outcome_label.grid(row=0, column=2, padx=10, pady=5, sticky="w")

        # Clear the lastline label
        self.lastline_label.config(text="")

        # Update the lastline label with final test results
        if self.stats.test_session_data:
            lastline = self.stats.test_session_data.lastline
            self.lastline_label.config(text=lastline)

    def start_file_monitoring(self):
        if self.file_path is not None and self.file_path.is_file():
            event_handler = FileChangeEventHandler(self.file_changed_callback)
            event_handler.file_path = (
                self.file_path
            )  # Set the file_path attribute in the event handler
            self.file_observer = Observer()
            self.file_observer.schedule(
                event_handler, str(self.file_path.parent), recursive=False
            )
            self.file_observer.start()
        else:
            print("Invalid file path.")

    def file_changed_callback(self):
        self.fetch_results()

    def stop_file_monitoring(self):
        if self.file_observer:
            self.file_observer.stop()
            self.file_observer.join()

    def __del__(self):
        self.stop_file_monitoring()

    def resize_window(self):
        new_height = min(
            TERM_SIZE.height - 200, self.stats.test_session_data.tot_num_to_run
        )
        self.root.geometry(f"{APP_WIDTH}x{new_height}")


def main():
    root = tk.Tk()
    root.geometry(f"{APP_WIDTH}x{APP_HEIGHT}")
    TestResultsGUI(root)
    root.mainloop()


if __name__ == "__main__":
    main()
